
/*************************************************************************
 * This unit implements an O(|M|) time algorithm described in            *
 *    K. Lemstrm and V. Mkinen: On Minimum Splitting of Pattern        *
 *    in Multi-Track String Matching, Accepted to CPM'2003,              *
 *    Morelia, Mexico, June 25-27, 2003.                                 *
 * Here M={(i,j,k) | p_i=t^k_j}, that is the set of matching character   *
 * pairs between pattern string P and each text T^k, 1<=k<=K.            *
 * The algorithm tries to find a splitting of P into kappa pieces so     *
 * that each piece is found in some text track T^k, and the occurrences  *
 * of the pieces are consecutive having at most an alpha-gap between     *
 * two pieces. The objective is to minimize kappa.                       *
 *                                                                       *
 * Copyright Veli Mkinen, Univ. Helsinki, Dept. CS, March 2003          *
 *    vmakinen@cs.helsinki.fi                                            *
 *************************************************************************/

// To handle alpha-gaps, we use a linear time sliding window minima algorithm
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <time.h>
#include "slidemin.c"

#define max2(a,b) ((a)>(b)?(a):(b))
#define min2(a,b) ((a)<(b)?(a):(b))

// Let's fix some values for demonstration
int m = 10;    // length of pattern
int n = 10000; // length of multi-track text
int K = 15;    // number of tracks
int alpha=5;   // maximum gap

unsigned char *P;  // pattern, P_{1...m}
unsigned char **T; // multi-track text, an array of texts T[k]_{1...n}

typedef struct {
   tripleNode *first;
   tripleNode *last;
} matchList ;

tripleNode  *newtripleNode()
{
   tripleNode *node = (tripleNode*)malloc(sizeof(tripleNode));
   // correct values for first row!
   node->next = NULL; node->prevD = NULL; node->kappa = 0;
   return node;
}

tripleNode  *newtripleNode2(int i, int j, int k, int kappa, tripleNode *next, tripleNode *prevD)
{
   tripleNode *node = (tripleNode*)malloc(sizeof(tripleNode));
   node->i = i; node->j = j; node->k = k; node->kappa = kappa; node->next = next;
   node->prevD = prevD;
   return node;
}

tripleNode *rowCopy(tripleNode *node, int i)
{
   return newtripleNode2(i,node->j,node->k,node->kappa,node->next,node->prevD);
};


/* Assuming integer alphabet Sigma,
   each symbol keeps a list of its occurrences in multi-track text T */
static matchList Sigma[256];

tripleNode *Process(unsigned char *P,unsigned char **T, int m, int n, int K) {
   tripleNode *temp, *temp2, *optTrace;
   matchList *row = (matchList*)malloc(sizeof(matchList)*(m+1));
   TCartesianTree *CT = newCartesianTree();
   keyType key;
   int kappa;

   /***********************************************************************
    * this produces the match set M in row-by-row order,                  *
    * and does the computation on-the-fly                                 *
    ***********************************************************************/

   int i,j,k,s;
   for (s=0;s<256;s++) {
      Sigma[s].first = NULL;
      Sigma[s].last = NULL;
   }

   // the following collects the list of occurrences of each character in T
   for (j=1;j<=n;j++)
      for (k=1;k<=K;k++) {
         temp = newtripleNode();
         temp->next = NULL;
         temp->j = j;
         temp->k = k;
         if (Sigma[(unsigned int)T[k][j]].last != NULL) {
            Sigma[(unsigned int)T[k][j]].last->next = temp;
            Sigma[(unsigned int)T[k][j]].last = temp;
         }
         else {
            Sigma[(unsigned int)T[k][j]].first = temp;
            Sigma[(unsigned int)T[k][j]].last = temp;
         }
      }

   // the following constructs the match set for each row separately
   for (i=1;i<=m;i++) {
      temp = Sigma[(unsigned int)P[i]].first;
      if (temp == NULL) return NULL; // no occurrences possible
      row[i].first = rowCopy(temp,i);
      row[i].last = row[i].first;
      temp = temp->next;
      while (temp!=NULL) {
         row[i].last->next = rowCopy(temp,i);
         row[i].last = row[i].last->next;
         temp = temp->next;
      }
   }

   // let's assign the prevD pointers by simulating the
   // merging of each two consecutive rows
   for (i=2;i<=m;i++) {
      temp = row[i-1].first;
      temp2 = row[i].first;
      while (temp!=NULL && temp2!=NULL) {
         if (temp->j == temp2->j-1 && temp->k == temp2->k) {
            temp2->prevD = temp;
            temp = temp->next;
            temp2 = temp2->next;
         }
         else if (temp->j < temp2->j-1 || (temp->j == temp2->j-1 && temp->k < temp2->k))
            temp = temp->next;
         else
            temp2 = temp2->next;
      }
   }

   // Ready to do the computation!
   // Have to make similar simulation of merging as above to sweep
   // through two consecutive rows simultaneously.
   // Also sliding window minima need to be maintained.

   for (i=2;i<=m;i++) {
      temp = row[i-1].first;
      temp2 = row[i].first;
      while (temp2!=NULL) {
         // add values to min-queue
         while (temp != NULL && temp->j < temp2->j) {
            //key.x = temp->j; key.y = temp->k;
            key = temp;
            Push(CT,temp->kappa,key);
            temp = temp->next;
         }
         // remove value from min-queue
         while (!isEmpty(CT) && firstKey(CT)->j<temp2->j-alpha-1)
            Eject(CT);
         // now, value of temp2->kappa is minimum of temp2->prevD->kappa and
         // CT->findMin()+1
         if (!isEmpty(CT)) {
            temp2->kappa = findMin(CT)+1;
            temp2->prevTrace = findKeyOfMin(CT);
         }
         else temp2->kappa = m+1; // no splitting up to this point
         if (temp2->prevD!=NULL && temp2->prevD->kappa<temp2->kappa) {
            temp2->kappa = temp2->prevD->kappa;
            temp2->prevTrace = temp2->prevD;
         }
         temp2 = temp2->next;
      }
      Empty(CT);
   }

   // find the smallest kappa at the last row
   temp = row[m].first;
   optTrace = NULL;
   kappa = m+1;
   while (temp!=NULL) {
      if (temp->kappa < kappa) {kappa = temp->kappa; optTrace = temp;};
      temp = temp->next;
   }

   // let's empty all data structures
   free(CT);

/* These should be emptied only after the optimal path is extracted
   for (i=1;i<=m;i++) {
      temp = row[i].first;
      while (temp != NULL) {
         temp2 = temp;
         temp = temp->next;
         free(temp2);
      }
   }
*/
   free(row);
   for (s=0;s<256;s++) {
      temp = Sigma[s].first;
      while (temp != NULL) {
         temp2 = temp;
         temp = temp->next;
         free(temp2);
      }
   }

   return optTrace; //containing min kappa
}

int main(int argc, char **argv)
{
   int i,j,k,kappa;
   tripleNode *node;
   time_t t;

   if (argc>1)
      alpha = atoi(argv[1]);
   T = (unsigned char**)malloc(sizeof(unsigned char*)*(K+1));
   for (k=1;k<=K;k++)
      T[k] = (unsigned char*)malloc(sizeof(unsigned char)*(n+1));

   P = (unsigned char*)malloc(sizeof(unsigned char)*(m+1));

   srand((unsigned) time(&t));
   for (i=1;i<=m;i++)
      P[i] = (unsigned char)(rand() % 15);

   for (k=1;k<=K;k++)
      for (j=1;j<=n;j++)
         T[k][j] = (unsigned char)(rand() % 15);

  node = Process(P,T,m,n,K);

  if (node == NULL)
     printf("No occurrences found\n");
  else
     printf("Found an occurrence with %d splits\n",node->kappa);
  // Optimal path can be extracted from node by following
  // node->prevTrace->prevTrace->...
}





